/*
 *  soapUI, copyright (C) 2004-2007 eviware.com 
 *
 *  soapUI is free software; you can redistribute it and/or modify it under the 
 *  terms of version 2.1 of the GNU Lesser General Public License as published by 
 *  the Free Software Foundation.
 *
 *  soapUI is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
 *  even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 *  See the GNU Lesser General Public License for more details at gnu.org.
 */

package com.eviware.soapui.impl;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import javax.swing.ImageIcon;

import org.apache.log4j.Logger;
import org.apache.xmlbeans.XmlException;
import org.apache.xmlbeans.XmlOptions;

import com.eviware.soapui.SoapUI;
import com.eviware.soapui.config.SoapuiWorkspaceDocumentConfig;
import com.eviware.soapui.config.WorkspaceProjectConfig;
import com.eviware.soapui.config.WorkspaceProjectConfig.Status;
import com.eviware.soapui.config.WorkspaceProjectConfig.Type;
import com.eviware.soapui.impl.settings.XmlBeansSettingsImpl;
import com.eviware.soapui.impl.wsdl.WsdlProject;
import com.eviware.soapui.model.ModelItem;
import com.eviware.soapui.model.project.Project;
import com.eviware.soapui.model.settings.Settings;
import com.eviware.soapui.model.support.AbstractModelItem;
import com.eviware.soapui.model.workspace.Workspace;
import com.eviware.soapui.model.workspace.WorkspaceListener;
import com.eviware.soapui.settings.UISettings;
import com.eviware.soapui.support.SoapUIException;
import com.eviware.soapui.support.UISupport;

/**
 * Default Workspace implementation
 * 
 * @author Ole.Matzura
 */

public class WorkspaceImpl extends AbstractModelItem implements Workspace
{
	private final static Logger log = Logger.getLogger( WorkspaceImpl.class );
	private List<Project> projectList = new ArrayList<Project>();
	private SoapuiWorkspaceDocumentConfig workspaceConfig;
	private String path = null;
	private Set<WorkspaceListener> listeners = new HashSet<WorkspaceListener>();
	private ImageIcon workspaceIcon;
	private XmlBeansSettingsImpl settings;

	public WorkspaceImpl( String path ) throws XmlException, IOException
	{
		File file = new File( path );
		this.path = file.getAbsolutePath();
		loadWorkspace( file );
		workspaceIcon = UISupport.createImageIcon( "/workspace.gif" );
	}

	public void switchWorkspace( File file ) throws SoapUIException
	{
		// check first if valid workspace file
		if( file.exists() )
		{
			try
			{
				SoapuiWorkspaceDocumentConfig.Factory.parse( file );
			}
			catch( Exception e )
			{
				throw new SoapUIException( "Failed to load workspace: " + e.toString() );
			}
		}
		
		fireWorkspaceSwitching();
		
		while( projectList.size() > 0 )
		{
			Project project = projectList.remove( 0 );
			try
			{
				fireProjectRemoved( project );
			}
			finally
			{
				project.release();
			}
		}
		
		try
		{
			String oldName = getName();
		
			loadWorkspace( file );
			this.path = file.getAbsolutePath();
			
			for( Project project : projectList )
			{
				fireProjectAdded( project );
			}
			
			notifyPropertyChanged( ModelItem.NAME_PROPERTY, oldName, getName() );
		}
		catch(Exception e )
		{
			SoapUI.logError( e );
		}
		
		fireWorkspaceSwitched();
	}
	
	private void loadWorkspace( File file ) throws XmlException, IOException
	{
		if( file.exists() )
		{
			log.info( "Loading workspace from [" + file.getAbsolutePath() + "]" );
			workspaceConfig = SoapuiWorkspaceDocumentConfig.Factory.parse( file );
			if( workspaceConfig.getSoapuiWorkspace().getSettings() == null )
				workspaceConfig.getSoapuiWorkspace().addNewSettings();

			settings = new XmlBeansSettingsImpl( this, SoapUI.getSettings(), workspaceConfig
						.getSoapuiWorkspace().getSettings() );

			boolean closeOnStartup = getSettings().getBoolean( UISettings.CLOSE_PROJECTS );
			List<WorkspaceProjectConfig> projects = workspaceConfig.getSoapuiWorkspace().getProjectList();
			for( int i = 0; i < projects.size(); i++ )
			{
				WorkspaceProjectConfig wsc = projects.get( i );
				String str = wsc.getStringValue();

				try
				{
					WsdlProject project = new WsdlProject( str, this, false, !closeOnStartup &&
								wsc.getStatus() != Status.CLOSED && wsc.getType() != Type.REMOTE, wsc.getName() );
					
					projectList.add( project );
				}
				catch( Exception e )
				{
					UISupport.showErrorMessage( "Failed to load project [" + str
								+ "]\nin workspace; " + e.getMessage() );

					SoapUI.logError( e );
				}
			}
		}
		else
		{
			workspaceConfig = SoapuiWorkspaceDocumentConfig.Factory.newInstance();
			workspaceConfig.addNewSoapuiWorkspace().setName( "Projects" );
			workspaceConfig.getSoapuiWorkspace().addNewSettings();

			settings = new XmlBeansSettingsImpl( this, SoapUI.getSettings(), workspaceConfig
						.getSoapuiWorkspace().getSettings() );
		}
	}

	public void setPath( String path )
	{
		this.path = path;
	}

	public Project[] getProjects()
	{
		return projectList.toArray( new Project[projectList.size()] );
	}

	public void setName( String name )
	{
		String oldName = getName();
		
		workspaceConfig.getSoapuiWorkspace().setName( name );
		notifyPropertyChanged( ModelItem.NAME_PROPERTY, oldName, name );
	}
	
	public void setDescription( String description )
	{
		String oldDescription = getDescription();
		
		workspaceConfig.getSoapuiWorkspace().setDescription( description );
		notifyPropertyChanged( ModelItem.DESCRIPTION_PROPERTY, oldDescription, description );
	}

	public String getName()
	{
		return workspaceConfig.getSoapuiWorkspace().isSetName() ? workspaceConfig.getSoapuiWorkspace().getName() : "Projects";
	}

	public Project getProjectAt( int index )
	{
		return projectList.get( index );
	}

	public Project getProjectByName( String projectName )
	{
		for( Project project : projectList )
		{
			if( project.getName().equals( projectName ) )
				return project;
		}

		return null;
	}

	public int getProjectCount()
	{
		return projectList.size();
	}

	public void onClose()
	{
		save( false );
	}

	public void save( boolean workspaceOnly )
	{
		save( workspaceOnly, false );
	}
	
	public void save( boolean workspaceOnly, boolean skipProjectsWithRunningTests )
	{
		try
		{
			List<WorkspaceProjectConfig> projects = new ArrayList<WorkspaceProjectConfig>();

			// save projects first
			for( int c = 0; c < getProjectCount(); c++ )
			{
				WsdlProject project = ( WsdlProject ) getProjectAt( c );

				if( !workspaceOnly  )
				{
					if( skipProjectsWithRunningTests && SoapUI.getTestMonitor().hasRunningTests( project ))
					{
						log.warn( "Project [" + project.getName() + "] has running tests.. skipping save" );
					}
					else
					{
						String path = project.getPath();
						if( path == null )
						{
							if( UISupport.confirm( "Project [" + project.getName() + "] has not been saved, save now?", "Save Project" ))
							{
								project.save();
							}
						}
						else
						{
							project.save();
						}
					}
				}

				String path = project.getPath();
				if( path != null )
				{
					WorkspaceProjectConfig wpc = WorkspaceProjectConfig.Factory.newInstance();
					wpc.setStringValue( path );
					if( project.isRemote() )
						wpc.setType( Type.REMOTE );
					
					if( !project.isOpen() )
						wpc.setStatus( Status.CLOSED );
					
					wpc.setName( project.getName() );
					projects.add( wpc );
				}
			}

			if( path == null )
			{
				File file = UISupport.getFileDialogs().saveAs( this, "Save workspace", ".xml",
							"XML Files (*.xml)", null );
				if( file == null )
					return;

				path = file.getAbsolutePath();
			}

			workspaceConfig.getSoapuiWorkspace().setProjectArray(
						projects.toArray( new WorkspaceProjectConfig[projects.size()] ) );
			workspaceConfig.getSoapuiWorkspace().setSoapuiVersion( SoapUI.SOAPUI_VERSION );

			File workspaceFile = new File( path );
			workspaceConfig.save( workspaceFile, new XmlOptions().setSavePrettyPrint() );

			log.info( "Saved workspace to [" + workspaceFile.getAbsolutePath() + "]" );
		}
		catch( IOException e )
		{
			log.error( "Failed to save workspace: " + e.getMessage(), e );
		}
	}

	public void addWorkspaceListener( WorkspaceListener listener )
	{
		listeners.add( listener );
	}

	public void removeWorkspaceListener( WorkspaceListener listener )
	{
		listeners.remove( listener );
	}

	public Project importProject( String fileName ) throws SoapUIException
	{
		File projectFile = new File( fileName );

		WsdlProject project = new WsdlProject( projectFile.getAbsolutePath(), this );
		projectList.add( project );
		fireProjectAdded( project );

		save( true );

		return project;
	}

	
	public WsdlProject createProject( String name ) throws SoapUIException
	{
		File projectFile = new File( createProjectFileName( name ) );
		File file = UISupport.getFileDialogs().saveAs( this, "Create Project", ".xml",
					"XML Files (*.xml)", projectFile );
		if( file == null )
			return null;

		return ( WsdlProject ) createProject( name, file );
	}

	public WsdlProject createProject( String name, File file ) throws SoapUIException
	{
		File projectFile = file;
		while( projectFile != null && projectFile.exists() )
		{
			Boolean result = Boolean.FALSE;
			while( !result.booleanValue() )
			{
				result = UISupport.confirmOrCancel( "Project File exists, overwrite?", "Overwrite Project?" );
				if( result == null )
					return null;
				if( result.booleanValue() )
				{
					projectFile.delete();
				}
				else
				{
					projectFile = UISupport.getFileDialogs().saveAs( this, "Create Project", ".xml",
								"XML Files (*.xml)", projectFile );
					if( projectFile == null )
						return null;
					else
						break;
				}
			}
		}

		WsdlProject project = new WsdlProject( projectFile == null ? null : projectFile.getAbsolutePath(), this );

		project.setName( name );
		projectList.add( project );

		fireProjectAdded( project );

		try
		{
			if( projectFile != null )
				project.save();
		}
		catch( IOException e )
		{
			log.error( "Failed to save project: " + e.getMessage(), e );
		}
//		save( true );

		return project;
	}

	private void fireProjectAdded( Project project )
	{
		for( Iterator<WorkspaceListener> iter = listeners.iterator(); iter.hasNext(); )
		{
			WorkspaceListener listener = iter.next();
			listener.projectAdded( project );
		}
	}
	
	private void fireWorkspaceSwitching()
	{
		for( Iterator<WorkspaceListener> iter = listeners.iterator(); iter.hasNext(); )
		{
			WorkspaceListener listener = iter.next();
			listener.workspaceSwitching( this );
		}
	}
	
	private void fireWorkspaceSwitched()
	{
		for( Iterator<WorkspaceListener> iter = listeners.iterator(); iter.hasNext(); )
		{
			WorkspaceListener listener = iter.next();
			listener.workspaceSwitched( this );
		}
	}

	private String createProjectFileName( String name )
	{
		return name + "-soapui-project.xml";
	}

	public void removeProject( Project project )
	{
		int ix = projectList.indexOf( project );
		if( ix == -1 )
			throw new RuntimeException( "Project [" + project.getName()
						+ "] not available in workspace for removal" );

		projectList.remove( ix );

		try
		{
			fireProjectRemoved( project );
		}
		finally
		{
			project.release();
//			workspaceConfig.getSoapuiWorkspace().removeProject( ix );
		}
	}
	
	public Project reloadProject( Project project ) throws SoapUIException
	{
		int ix = projectList.indexOf( project );
		if( ix == -1 )
			throw new RuntimeException( "Project [" + project.getName()
						+ "] not available in workspace for reload" );

		projectList.remove( ix );
		fireProjectRemoved( project );
		
		String tempName = project.getName();
		project.release();
		project = new WsdlProject( project.getPath(), this, false, true, tempName );
		
		projectList.add( ix, project );

		fireProjectAdded( project );

//		workspaceConfig.getSoapuiWorkspace().getProjectArray( ix ).setStringValue( project.getPath() );
		
		return project;
	}

	private void fireProjectRemoved( Project project )
	{
		WorkspaceListener[] listenerArray = listeners
					.toArray( new WorkspaceListener[listeners.size()] );
		for( int c = 0; c < listenerArray.length; c++ )
		{
			listenerArray[c].projectRemoved( project );
		}
	}

	public ImageIcon getIcon()
	{
		return workspaceIcon;
	}

	public Settings getSettings()
	{
		return settings;
	}

	public int getIndexOfProject( Project project )
	{
		return projectList.indexOf( project );
	}

	public String getPath()
	{
		return path;
	}

	public void release()
	{
		settings.release();

		for( Project project : projectList )
			project.release();
	}

	public List<? extends Project> getProjectList()
	{
		return projectList;
	}

	public String getDescription()
	{
		return workspaceConfig.getSoapuiWorkspace().getDescription();
	}

	public WsdlProject importRemoteProject( String url ) throws SoapUIException
	{
		WsdlProject project = new WsdlProject( url, this, false );
		projectList.add( project );
		fireProjectAdded( project );

		save( true );

		return project;
	}

	public void closeProject( Project project )
	{
		int ix = projectList.indexOf( project );
		if( ix == -1 )
			throw new RuntimeException( "Project [" + project.getName()
						+ "] not available in workspace for close" );

		projectList.remove( ix );
		fireProjectRemoved( project );
		
		String name = project.getName();
		project.release();
		
		try
		{
			project = new WsdlProject( project.getPath(), this, false, false, name );
			projectList.add( ix, project );
			fireProjectAdded( project );
		}
		catch( Exception e )
		{
			UISupport.showErrorMessage( "Failed to close project [" + name
						+ "]\nin workspace; " + e.getMessage() );

			SoapUI.logError( e );
		}
	}

	public List<Project> getOpenProjectList()
	{
		List<Project> availableProjects = new ArrayList<Project>(); 
		
		for( Project project : projectList )
			if( project.isOpen() )
				availableProjects.add( project );
		
		return availableProjects;
	}

	public Project openProject( Project project ) throws SoapUIException
	{
		return reloadProject( project );
	}
}
