/*
 *  soapUI, copyright (C) 2004-2007 eviware.com 
 *
 *  soapUI is free software; you can redistribute it and/or modify it under the 
 *  terms of version 2.1 of the GNU Lesser General Public License as published by 
 *  the Free Software Foundation.
 *
 *  soapUI is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
 *  even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 *  See the GNU Lesser General Public License for more details at gnu.org.
 */

package com.eviware.soapui.impl.wsdl;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import com.eviware.soapui.SoapUI;
import com.eviware.soapui.config.LoadTestConfig;
import com.eviware.soapui.config.TestCaseConfig;
import com.eviware.soapui.config.TestSuiteConfig;
import com.eviware.soapui.config.TestSuiteRunTypesConfig;
import com.eviware.soapui.config.TestSuiteRunTypesConfig.Enum;
import com.eviware.soapui.impl.wsdl.loadtest.WsdlLoadTest;
import com.eviware.soapui.impl.wsdl.testcase.WsdlTestCase;
import com.eviware.soapui.impl.wsdl.teststeps.WsdlTestStep;
import com.eviware.soapui.model.testsuite.TestCase;
import com.eviware.soapui.model.testsuite.TestSuite;
import com.eviware.soapui.model.testsuite.TestSuiteListener;

/**
 * TestSuite implementation for WSDL projects.
 * 
 * @author Ole.Matzura
 */

public class WsdlTestSuite extends AbstractWsdlModelItem<TestSuiteConfig> implements TestSuite
{
   private final WsdlProject project;
   private List<WsdlTestCase> testCases = new ArrayList<WsdlTestCase>();
   private Set<TestSuiteListener> testSuiteListeners = new HashSet<TestSuiteListener>();
   
   public WsdlTestSuite(WsdlProject project, TestSuiteConfig config)
   {
   	super( config, project, "/testSuite.gif" );
      this.project = project;
      
      List<TestCaseConfig> testCaseConfigs = config.getTestCaseList();
      for (int i = 0; i < testCaseConfigs.size(); i++)
      {
         testCases.add( new WsdlTestCase( this, testCaseConfigs.get(i), false ));
      }
      
      if( !config.isSetRunType() )
      	config.setRunType( TestSuiteRunTypesConfig.SEQUENTIAL );
      
   	for( TestSuiteListener listener : SoapUI.getListenerRegistry().getListeners( TestSuiteListener.class ) )
		{
			addTestSuiteListener( listener );
		}
   }
   
	public TestSuiteRunType getRunType()
	{
		Enum runType = getConfig().getRunType();
		
		if( runType.equals( TestSuiteRunTypesConfig.PARALLELL ))
			return TestSuiteRunType.PARALLEL;
		else
			return TestSuiteRunType.SEQUENTIAL;
	}
	
	public void setRunType( TestSuiteRunType runType )
	{
		TestSuiteRunType oldRunType = getRunType();
		
		if( runType == TestSuiteRunType.PARALLEL && oldRunType != TestSuiteRunType.PARALLEL )
		{
			getConfig().setRunType( TestSuiteRunTypesConfig.PARALLELL );
			notifyPropertyChanged( RUNTYPE_PROPERTY, oldRunType, runType );
		}
		else if( runType == TestSuiteRunType.SEQUENTIAL && oldRunType != TestSuiteRunType.SEQUENTIAL )
		{
			getConfig().setRunType( TestSuiteRunTypesConfig.SEQUENTIAL );
			notifyPropertyChanged( RUNTYPE_PROPERTY, oldRunType, runType );
		}
	}

	public WsdlProject getProject()
   {
      return project;
   }

   public int getTestCaseCount()
   {
      return testCases.size();
   }

   public WsdlTestCase getTestCaseAt(int index)
   {
      return testCases.get( index );
   }

   public WsdlTestCase getTestCaseByName(String testCaseName)
	{
		return ( WsdlTestCase ) getWsdlModelItemByName( testCases, testCaseName );
	}

	public WsdlTestCase cloneTestCase( WsdlTestCase testCase, String name )
   {
		testCase.onSave();
   	TestCaseConfig newTestCase = getConfig().addNewTestCase();
		newTestCase.set( testCase.getConfig() );
		newTestCase.setName( name );
		WsdlTestCase newWsdlTestCase = new WsdlTestCase( this, newTestCase, false );
		
      testCases.add( newWsdlTestCase );
      fireTestCaseAdded( newWsdlTestCase );

      return newWsdlTestCase;
   }
   
   public WsdlTestCase addNewTestCase( String name )
   {
      WsdlTestCase testCase = new WsdlTestCase( this, getConfig().addNewTestCase(), false );
      testCase.setName( name );
      testCase.setFailOnError( true );
      testCase.setSearchProperties( true );
      testCases.add( testCase );
      fireTestCaseAdded( testCase );

      return testCase;
   }
   
   public WsdlTestCase importTestCase( WsdlTestCase testCase, String name, int index, boolean includeLoadTests )
   {
   	testCase.onSave();
   	 
   	if( index >= testCases.size() )
   		index = -1; //testCases.size()-1;
   	
      TestCaseConfig testCaseConfig = index == -1 ? 
      	( TestCaseConfig ) getConfig().addNewTestCase().set( testCase.getConfig().copy() ) :
         ( TestCaseConfig ) getConfig().insertNewTestCase( index ).set( testCase.getConfig().copy() );
      testCaseConfig.setName( name );
      
      if( !includeLoadTests )
      	testCaseConfig.setLoadTestArray( new LoadTestConfig[0] );
     
		testCase = new WsdlTestCase( this, testCaseConfig, false );
		
		if( index == -1 )
			testCases.add( testCase );
		else
			testCases.add( index, testCase );
		
      fireTestCaseAdded( testCase );

      return testCase;
   }
   
   public void removeTestCase(WsdlTestCase testCase )
   {
      int ix = testCases.indexOf( testCase );

      testCases.remove( ix );
      try
      {
      	fireTestCaseRemoved( testCase );
      }
      finally
      {
      	testCase.release();
      	getConfig().removeTestCase( ix );
      }
   }

   public void fireTestCaseAdded( WsdlTestCase testCase )
   {
      TestSuiteListener[] a = testSuiteListeners.toArray( new TestSuiteListener[testSuiteListeners.size()] );
      
      for (int c = 0; c < a.length; c++ )
      {
         a[c].testCaseAdded( testCase );
      }
   }
   
   public void fireTestCaseRemoved( WsdlTestCase testCase )
   {
   	TestSuiteListener[] a = testSuiteListeners.toArray( new TestSuiteListener[testSuiteListeners.size()] );
      
      for (int c = 0; c < a.length; c++ )
      {
         a[c].testCaseRemoved( testCase );
      }
   } 
   
   private void fireTestCaseMoved( WsdlTestCase testCase, int ix, int offset )
	{
   	TestSuiteListener[] a = testSuiteListeners.toArray( new TestSuiteListener[testSuiteListeners.size()] );
      
      for (int c = 0; c < a.length; c++ )
      {
         a[c].testCaseMoved( testCase, ix, offset );
      }
	}

   public void fireTestStepAdded( WsdlTestStep testStep, int index )
   {
   	TestSuiteListener[] a = testSuiteListeners.toArray( new TestSuiteListener[testSuiteListeners.size()] );
      
      for (int c = 0; c < a.length; c++ )
      {
         a[c].testStepAdded( testStep, index );
      }
   }
   
   public void fireTestStepRemoved( WsdlTestStep testStep, int ix )
   {
   	TestSuiteListener[] a = testSuiteListeners.toArray( new TestSuiteListener[testSuiteListeners.size()] );
      
      for (int c = 0; c < a.length; c++ )
      {
         a[c].testStepRemoved( testStep, ix );
      }
   } 
   
   public void fireTestStepMoved( WsdlTestStep testStep, int ix, int offset )
   {
   	TestSuiteListener[] a = testSuiteListeners.toArray( new TestSuiteListener[testSuiteListeners.size()] );
      
      for (int c = 0; c < a.length; c++ )
      {
         a[c].testStepMoved( testStep, ix, offset );
      }
   } 

   public void fireLoadTestAdded( WsdlLoadTest loadTest )
   {
   	TestSuiteListener[] a = testSuiteListeners.toArray( new TestSuiteListener[testSuiteListeners.size()] );
      
      for (int c = 0; c < a.length; c++ )
      {
         a[c].loadTestAdded( loadTest );
      }
   }
   
   public void fireLoadTestRemoved( WsdlLoadTest loadTest )
   {
   	TestSuiteListener[] a = testSuiteListeners.toArray( new TestSuiteListener[testSuiteListeners.size()] );
      
      for (int c = 0; c < a.length; c++ )
      {
         a[c].loadTestRemoved( loadTest );
      }
   } 
   
	public void addTestSuiteListener(TestSuiteListener listener)
	{
		testSuiteListeners.add( listener );
	}

	public void removeTestSuiteListener(TestSuiteListener listener)
	{
		testSuiteListeners.remove( listener );
	}

	public int getTestCaseIndex(TestCase testCase)
	{
		return testCases.indexOf( testCase );
	}

	public void release()
	{
		super.release();
		
		for( WsdlTestCase testCase : testCases )
			testCase.release();
		
		testSuiteListeners.clear();
	}

	public List<TestCase> getTestCaseList()
	{
		List<TestCase> result = new ArrayList<TestCase>();
		for( WsdlTestCase testCase : testCases )
			result.add( testCase );
		
		return result;
	} 
	
	/**
	 * Moves a testcase by the specified offset, a bit awkward since xmlbeans doesn't support reordering
	 * of arrays, we need to create copies of the contained XmlObjects
	 * 
	 * @param ix
	 * @param offset
	 */
	
	public WsdlTestCase moveTestCase(int ix, int offset)
	{
		WsdlTestCase testCase = testCases.get( ix );

		if( offset == 0 ) 
			return testCase;

		testCases.remove( ix );
		testCases.add( ix+offset, testCase );

		TestCaseConfig [] configs = new TestCaseConfig[testCases.size()];
		
		for( int c = 0; c < testCases.size(); c++ )
		{
			if( offset > 0 )
			{
				if( c < ix )
					configs[c] = (TestCaseConfig) getConfig().getTestCaseArray(c).copy();
				else if( c < (ix+offset))
					configs[c] = (TestCaseConfig) getConfig().getTestCaseArray(c+1).copy();
				else if( c == ix+offset )
					configs[c] = (TestCaseConfig) getConfig().getTestCaseArray(ix).copy();
				else
					configs[c] = (TestCaseConfig) getConfig().getTestCaseArray(c).copy();
			}
			else
			{
				if( c < ix+offset )
					configs[c] = (TestCaseConfig) getConfig().getTestCaseArray(c).copy();
				else if( c == ix+offset )
					configs[c] = (TestCaseConfig) getConfig().getTestCaseArray(ix).copy();
				else if( c <= ix )
					configs[c] = (TestCaseConfig) getConfig().getTestCaseArray(c-1).copy();
				else
					configs[c] = (TestCaseConfig) getConfig().getTestCaseArray(c).copy();				
			}
		}
		
		getConfig().setTestCaseArray( configs );
		for( int c = 0; c < configs.length; c++ )
		{
			testCases.get( c ).resetConfigOnMove( getConfig().getTestCaseArray( c ) );
		}
		
		fireTestCaseMoved(testCase, ix, offset );
		return testCase;
	}

	public int getIndexOfTestCase( TestCase testCase )
	{
		return testCases.indexOf( testCase );
	}

	@Override
	public void onSave()
	{
		for( WsdlTestCase testCase : testCases )
			testCase.onSave();
	}
}
