/*
 *  soapUI, copyright (C) 2004-2007 eviware.com 
 *
 *  soapUI is free software; you can redistribute it and/or modify it under the 
 *  terms of version 2.1 of the GNU Lesser General Public License as published by 
 *  the Free Software Foundation.
 *
 *  soapUI is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
 *  even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 *  See the GNU Lesser General Public License for more details at gnu.org.
 */

package com.eviware.soapui.impl.wsdl.mock;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.swing.ImageIcon;

import com.eviware.soapui.SoapUI;
import com.eviware.soapui.config.MockOperationConfig;
import com.eviware.soapui.config.MockServiceConfig;
import com.eviware.soapui.impl.wsdl.AbstractWsdlModelItem;
import com.eviware.soapui.impl.wsdl.WsdlInterface;
import com.eviware.soapui.impl.wsdl.WsdlOperation;
import com.eviware.soapui.impl.wsdl.WsdlProject;
import com.eviware.soapui.impl.wsdl.support.ModelItemIconAnimator;
import com.eviware.soapui.impl.wsdl.testcase.WsdlTestRunContext;
import com.eviware.soapui.model.iface.Operation;
import com.eviware.soapui.model.mock.MockResult;
import com.eviware.soapui.model.mock.MockRunListener;
import com.eviware.soapui.model.mock.MockRunner;
import com.eviware.soapui.model.mock.MockService;
import com.eviware.soapui.model.mock.MockServiceListener;
import com.eviware.soapui.model.project.Project;

/**
 * A MockService for simulation WsdlInterfaces and their operations
 * 
 * @author ole.matzura
 */

public class WsdlMockService extends AbstractWsdlModelItem<MockServiceConfig> implements MockService
{
	private static final String REQUIRE_SOAP_VERSION = WsdlMockService.class.getName() + "@require-soap-version";
	private List<WsdlMockOperation> mockOperations = new ArrayList<WsdlMockOperation>();
   private Set<MockRunListener> mockRunListeners = new HashSet<MockRunListener>();
   private Set<MockServiceListener> mockServiceListeners = new HashSet<MockServiceListener>();
   private MockServiceIconAnimator iconAnimator;
	private WsdlMockRunner mockRunner;
	
	public WsdlMockService( Project project, MockServiceConfig config)
	{
		super( config, project, "/mockService.gif" );
		
		List<MockOperationConfig> testStepConfigs = config.getMockOperationList();
      for (MockOperationConfig tsc : testStepConfigs )
      {
         WsdlMockOperation testStep = new WsdlMockOperation( this, tsc );
         mockOperations.add( testStep );
      }
      
      if( !config.isSetPort() || config.getPort() < 1 )
      	config.setPort( 8080 );
      
      if( !config.isSetPath() )
      	config.setPath( "/" );
      
      iconAnimator = new MockServiceIconAnimator();
      addMockRunListener( iconAnimator );
      
      for (MockRunListener listener : SoapUI.getListenerRegistry().getListeners( MockRunListener.class ))
      {
          addMockRunListener(listener);
      }
	}

	public void addMockRunListener( MockRunListener listener )
	{
		mockRunListeners.add( listener );
	}

	public String getPath()
	{
		return getConfig().getPath();
	}

	public WsdlMockOperation getMockOperationAt( int index )
	{
		return mockOperations.get( index );
	}

	public WsdlMockOperation getMockOperationByName( String name )
	{
		return ( WsdlMockOperation ) getWsdlModelItemByName( mockOperations, name );
	}

	public int getMockOperationCount()
	{
		return mockOperations.size();
	}

	public WsdlProject getProject()
	{
		return ( WsdlProject ) getParent();
	}

	public int getPort()
	{
		return getConfig().getPort();
	}

	public void removeMockRunListener( MockRunListener listener )
	{
		mockRunListeners.remove( listener );
	}

	public WsdlMockRunner start( WsdlTestRunContext context ) throws Exception
	{
		String path = getPath();
		if( path == null || path.trim().length() == 0 || path.trim().charAt( 0 ) != '/' )
			throw new Exception( "Invalid path; must start with '/'");
		
		mockRunner = new WsdlMockRunner( this, context );
		return mockRunner;
	}

	public WsdlMockRunner getMockRunner()
	{
		return mockRunner;
	}

	public WsdlMockOperation getMockOperation( Operation operation )
	{
		for( int c = 0; c < getMockOperationCount(); c++ )
		{
			WsdlMockOperation mockOperation = mockOperations.get( c );
			if( mockOperation.getOperation() == operation )
				return mockOperation;
		}
		
		return null;
	}

	public WsdlMockOperation addNewMockOperation( WsdlOperation operation )
	{
		if( getMockOperation( operation ) != null )
			return null;
		
		MockOperationConfig config = getConfig().addNewMockOperation();
		config.setName( operation.getName() );
		WsdlMockOperation mockOperation = new WsdlMockOperation( this, config, operation );
		
		mockOperations.add( mockOperation );
		fireMockOperationAdded( mockOperation );
		
		return mockOperation;
	}

	public void setPort( int port )
	{
		String oldEndpoint = getLocalEndpoint();
		
		int oldPort = getPort();
		if( port != oldPort )
		{
			getConfig().setPort( port );
			notifyPropertyChanged( PORT_PROPERTY, oldPort, port );

			for( WsdlInterface iface : getMockedInterfaces())
			{
				if( Arrays.asList( iface.getEndpoints() ).contains( oldEndpoint ))
					iface.changeEndpoint( oldEndpoint, getLocalEndpoint() );
			}
		}
	}

	public WsdlInterface [] getMockedInterfaces()
	{
		Set<WsdlInterface> result = new HashSet<WsdlInterface>();
		
		for( WsdlMockOperation mockOperation : mockOperations )
		{
			WsdlOperation operation = mockOperation.getOperation();
			if( operation != null )
				result.add( operation.getInterface() );
		}
		
		return result.toArray( new WsdlInterface[result.size()] );
	}

	public void release()
	{
		super.release();
		
		for( WsdlMockOperation operation : mockOperations )
			operation.release();
		
		mockServiceListeners.clear();
	}
	
	public void setPath( String path )
	{
		String oldEndpoint = getLocalEndpoint();
		
		String oldPath = getPath();
		if( !path.equals( oldPath ))
		{
			getConfig().setPath( path );
			notifyPropertyChanged( PATH_PROPERTY, oldPath, path );
			
			for( WsdlInterface iface : getMockedInterfaces())
			{
				if( Arrays.asList( iface.getEndpoints() ).contains( oldEndpoint ))
					iface.changeEndpoint( oldEndpoint, getLocalEndpoint() );
			}
		}
	}

	public MockRunListener [] getMockRunListeners()
	{
		return mockRunListeners.toArray( new MockRunListener[mockRunListeners.size()] );
	}

	public void removeMockOperation( WsdlMockOperation mockOperation )
	{
		int ix = mockOperations.indexOf( mockOperation );
		if( ix == -1 )
			throw new RuntimeException( "Unkonws MockOperation specified to removeMockOperation" );
		
		mockOperations.remove( ix );
		fireMockOperationRemoved( mockOperation );
		mockOperation.release();
		getConfig().removeMockOperation( ix );
	}

	public void addMockServiceListener( MockServiceListener listener )
	{
		mockServiceListeners.add(  listener  );
	}

	public void removeMockServiceListener( MockServiceListener listener )
	{
		mockServiceListeners.remove( listener );
	}
	
	protected void fireMockOperationAdded( WsdlMockOperation mockOperation )
	{
		MockServiceListener[] listeners = mockServiceListeners.toArray( new MockServiceListener[mockServiceListeners.size()] );
		for( MockServiceListener listener : listeners )
		{
			listener.mockOperationAdded( mockOperation );
		}
	}
	
	protected void fireMockOperationRemoved( WsdlMockOperation mockOperation )
	{
		MockServiceListener[] listeners = mockServiceListeners.toArray( new MockServiceListener[mockServiceListeners.size()] );
		for( MockServiceListener listener : listeners )
		{
			listener.mockOperationRemoved( mockOperation );
		}
	}
	
	protected void fireMockResponseAdded( WsdlMockResponse mockResponse )
	{
		MockServiceListener[] listeners = mockServiceListeners.toArray( new MockServiceListener[mockServiceListeners.size()] );
		for( MockServiceListener listener : listeners )
		{
			listener.mockResponseAdded( mockResponse );
		}
	}
	
	protected void fireMockResponseRemoved( WsdlMockResponse mockResponse )
	{
		MockServiceListener[] listeners = mockServiceListeners.toArray( new MockServiceListener[mockServiceListeners.size()] );
		for( MockServiceListener listener : listeners )
		{
			listener.mockResponseRemoved( mockResponse );
		}
	}

	@Override
	public ImageIcon getIcon()
	{
		return iconAnimator.getIcon();
	}

	private class MockServiceIconAnimator extends ModelItemIconAnimator implements MockRunListener
	{
		public MockServiceIconAnimator()
		{
			super( WsdlMockService.this, "/mockService.gif", 
					new String[] {"/mockService_1.gif", "/mockService_2.gif",
						"/mockService_3.gif", "/mockService_4.gif"} );
		}

		public void onMockRequest( MockRunner runner, HttpServletRequest request, HttpServletResponse response )
		{
		}

		public void onMockResult( MockResult result )
		{
		}

		public void onMockRunnerStart( MockRunner mockRunner )
		{
			start();
		}

		public void onMockRunnerStop( MockRunner mockRunner )
		{
			stop();
			WsdlMockService.this.mockRunner = null; 
		}
	}
	
	public String getLocalEndpoint()
	{
		return "http://127.0.0.1:" + getPort() + getPath();
	}

	public boolean isRequireSoapVersion()
	{
		return getSettings().getBoolean( REQUIRE_SOAP_VERSION );
	}
	
	public void setRequireSoapVersion( boolean requireSoapVersion )
	{
	   getSettings().setBoolean( REQUIRE_SOAP_VERSION, requireSoapVersion );	
	}

	public WsdlMockRunner start() throws Exception
	{
		return start( null );
	}

	public boolean hasMockOperation( Operation operation )
	{
		return getMockOperation( operation ) != null;
	}

	@Override
	public void onSave()
	{
		for( WsdlMockOperation mockOperation : mockOperations )
			mockOperation.onSave();
	}
}
