/*
 *  soapUI, copyright (C) 2004-2007 eviware.com 
 *
 *  soapUI is free software; you can redistribute it and/or modify it under the 
 *  terms of version 2.1 of the GNU Lesser General Public License as published by 
 *  the Free Software Foundation.
 *
 *  soapUI is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
 *  even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 *  See the GNU Lesser General Public License for more details at gnu.org.
 */

package com.eviware.soapui.impl.wsdl.panels.request.components.editor.views.source;

import javax.swing.JPopupMenu;
import javax.swing.JSeparator;

import com.eviware.soapui.impl.wsdl.WsdlInterface;
import com.eviware.soapui.impl.wsdl.WsdlOperation;
import com.eviware.soapui.impl.wsdl.WsdlRequest;
import com.eviware.soapui.impl.wsdl.actions.request.AddWSSUsernameTokenAction;
import com.eviware.soapui.impl.wsdl.actions.request.AddWSTimestampAction;
import com.eviware.soapui.impl.wsdl.mock.WsdlMockResponse;
import com.eviware.soapui.impl.wsdl.panels.mockoperation.WsdlMockResponseMessageExchange;
import com.eviware.soapui.impl.wsdl.panels.mockoperation.WsdlMockResultMessageExchange;
import com.eviware.soapui.impl.wsdl.panels.request.components.editor.XmlEditor;
import com.eviware.soapui.impl.wsdl.panels.request.components.editor.XmlEditorView;
import com.eviware.soapui.impl.wsdl.panels.request.components.editor.support.ValidationError;
import com.eviware.soapui.impl.wsdl.panels.request.components.editor.views.registry.RequestEditorViewFactory;
import com.eviware.soapui.impl.wsdl.panels.request.components.editor.views.registry.ResponseEditorViewFactory;
import com.eviware.soapui.impl.wsdl.support.wsdl.WsdlValidator;
import com.eviware.soapui.impl.wsdl.testcase.WsdlTestRunContext;
import com.eviware.soapui.impl.wsdl.teststeps.WsdlResponseMessageExchange;
import com.eviware.soapui.impl.wsdl.teststeps.WsdlTestRequest;
import com.eviware.soapui.impl.wsdl.teststeps.assertions.AssertionError;
import com.eviware.soapui.model.ModelItem;
import com.eviware.soapui.support.xml.JXEditTextArea;

/**
 * Factory for default "XML" source editor view in soapUI
 * 
 * @author ole.matzura
 */

public class XmlSourceEditorFactory implements ResponseEditorViewFactory, RequestEditorViewFactory
{
	public static final String VIEW_ID = "Source";

	public XmlEditorView createEditorView(XmlEditor editor)
	{
		return new XmlSourceEditorView( editor );
	}

	public String getViewId()
	{
		return VIEW_ID;
	}

	public XmlEditorView createRequestEditorView( XmlEditor editor, ModelItem modelItem )
	{
		if( modelItem instanceof WsdlRequest )
		{
			return new WsdlRequestXmlSourceEditor( editor, ( WsdlRequest ) modelItem );
		}
		else if( modelItem instanceof WsdlMockResponse )
		{
			return new WsdlMockRequestXmlSourceEditor( editor, ( WsdlMockResponse ) modelItem ); 
		}
		
		return null;
	}
	
	public XmlEditorView createResponseEditorView( XmlEditor editor, ModelItem modelItem )
	{
		if( modelItem instanceof WsdlRequest )
		{
			return new WsdlResponseXmlSourceEditor( editor, ( WsdlRequest ) modelItem );
		}
		else if( modelItem instanceof WsdlMockResponse )
		{
			return new WsdlMockResponseXmlSourceEditor( editor, ( WsdlMockResponse ) modelItem ); 
		}
		
		return null;
	}
	
	/**
	 * XmlSource editor for a WsdlRequest
	 * 
	 * @author ole.matzura
	 */
	
	public static class WsdlRequestXmlSourceEditor extends XmlSourceEditorView
	{
		private final WsdlRequest request;

		public WsdlRequestXmlSourceEditor(XmlEditor xmlEditor, WsdlRequest request )
		{
			super(xmlEditor);
			this.request = request;
		}

		protected ValidationError[] validateXml( String xml )
		{
			WsdlOperation operation = (WsdlOperation) request.getOperation();
			WsdlValidator validator = new WsdlValidator(((WsdlInterface) operation.getInterface()).getWsdlContext());

			WsdlResponseMessageExchange wsdlResponseMessageExchange = new WsdlResponseMessageExchange( request );
			wsdlResponseMessageExchange.setRequestContent( xml );
			return validator.assertRequest( wsdlResponseMessageExchange, false );
		}

		protected void buildPopup(JPopupMenu inputPopup, JXEditTextArea editArea )
		{
			super.buildPopup(inputPopup, editArea );

			inputPopup.insert( new JSeparator(), 2 );
			inputPopup.insert(new AddWSSUsernameTokenAction(request),3);
			inputPopup.insert(new AddWSTimestampAction(request),4);
		}
	}
	
	/**
	 * XmlSource editor for a WsdlMockRequest
	 * 
	 * @author ole.matzura
	 */
	
	public static class WsdlMockRequestXmlSourceEditor extends XmlSourceEditorView
	{
		private final WsdlMockResponse mockResponse;

		public WsdlMockRequestXmlSourceEditor(XmlEditor xmlEditor, WsdlMockResponse mockResponse )
		{
			super(xmlEditor);
			this.mockResponse = mockResponse;
		}

		protected ValidationError[] validateXml( String xml )
		{
			WsdlOperation operation = mockResponse.getMockOperation().getOperation();
			
			if( operation == null )
			{
				return new ValidationError[] { new AssertionError( "Missing operation for MockResponse")};
			}
			
			WsdlValidator validator = new WsdlValidator(((WsdlInterface) operation.getInterface()).getWsdlContext());
			return validator.assertRequest( new WsdlMockResultMessageExchange( mockResponse.getMockResult() ), false );
		}

		protected void buildPopup(JPopupMenu inputPopup, JXEditTextArea editArea )
		{
			super.buildPopup(inputPopup, editArea );
			inputPopup.insert( new JSeparator(), 2 );
		}
	}
	
	/**
	 * XmlSource editor for a WsdlResponse
	 * 
	 * @author ole.matzura
	 */
	
	public static class WsdlResponseXmlSourceEditor extends XmlSourceEditorView
	{
		private final WsdlRequest request;

		public WsdlResponseXmlSourceEditor(XmlEditor xmlEditor, WsdlRequest request )
		{
			super(xmlEditor);
			this.request = request;
		}

		protected ValidationError[] validateXml( String xml )
		{
			if( request instanceof WsdlTestRequest )
			{
				WsdlTestRequest testRequest = (WsdlTestRequest)request;
				testRequest.assertResponse( new WsdlTestRunContext(  testRequest.getTestStep() ));
			}
			
			WsdlOperation operation = (WsdlOperation)request.getOperation();
			WsdlValidator validator = new WsdlValidator(((WsdlInterface) operation.getInterface()).getWsdlContext());

			return validator.assertResponse( new WsdlResponseMessageExchange( request ), false );
		}
	}
	
	/**
	 * XmlSource editor for a WsdlMockResponse
	 * 
	 * @author ole.matzura
	 */
	
	public static class WsdlMockResponseXmlSourceEditor extends XmlSourceEditorView
	{
		private final WsdlMockResponse mockResponse;

		public WsdlMockResponseXmlSourceEditor(XmlEditor xmlEditor, WsdlMockResponse mockResponse)
		{
			super(xmlEditor);
			this.mockResponse = mockResponse;
		}

		protected ValidationError[] validateXml( String xml )
		{
			WsdlOperation operation = mockResponse.getMockOperation().getOperation();
			if( operation == null )
			{
				return new ValidationError[] { new AssertionError( "Missing operation for MockResponse")};
			}
			
			WsdlValidator validator = new WsdlValidator(((WsdlInterface) operation.getInterface()).getWsdlContext());
			return validator.assertResponse( new WsdlMockResponseMessageExchange( mockResponse ), false );
		}
	}
}
