/*
 *  soapUI, copyright (C) 2004-2007 eviware.com 
 *
 *  soapUI is free software; you can redistribute it and/or modify it under the 
 *  terms of version 2.1 of the GNU Lesser General Public License as published by 
 *  the Free Software Foundation.
 *
 *  soapUI is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
 *  even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 *  See the GNU Lesser General Public License for more details at gnu.org.
 */

package com.eviware.soapui.impl.wsdl.teststeps;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Logger;
import org.apache.xmlbeans.XmlCursor;
import org.apache.xmlbeans.XmlException;
import org.apache.xmlbeans.XmlObject;
import org.apache.xmlbeans.XmlOptions;
import org.apache.xmlbeans.XmlCursor.TokenType;
import org.w3c.dom.DocumentFragment;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import com.eviware.soapui.config.ValueTransferConfig;
import com.eviware.soapui.impl.wsdl.submit.filters.PropertyExpansionRequestFilter;
import com.eviware.soapui.model.iface.SubmitContext;
import com.eviware.soapui.model.support.TestSuiteListenerAdapter;
import com.eviware.soapui.model.testsuite.TestCase;
import com.eviware.soapui.model.testsuite.TestStep;
import com.eviware.soapui.model.testsuite.TestStepProperty;
import com.eviware.soapui.support.PropertyChangeNotifier;
import com.eviware.soapui.support.xml.XmlUtils;

/**
 * Class for transferring a property value between 2 test steps. This class is relatively complex due
 * to backwards compatibility issues and to gracefull handling of references test steps and properties.
 * 
 * @author Ole.Matzura
 */

public class PropertyTransfer implements PropertyChangeNotifier
{
	private final static Logger log = Logger.getLogger( PropertyTransfer.class );
	
	public final static String SOURCE_PATH_PROPERTY = PropertyTransfer.class.getName() + "@sourcePath";
	public final static String SOURCE_TYPE_PROPERTY = PropertyTransfer.class.getName() + "@sourceProperty";
	public final static String SOURCE_STEP_PROPERTY = PropertyTransfer.class.getName() + "@sourceStep";
	public final static String TARGET_PATH_PROPERTY = PropertyTransfer.class.getName() + "@targetPath";
	public final static String TARGET_TYPE_PROPERTY = PropertyTransfer.class.getName() + "@targetProperty";
	public final static String TARGET_STEP_PROPERTY = PropertyTransfer.class.getName() + "@targetStep";
	public final static String NAME_PROPERTY = PropertyTransfer.class.getName() + "@name";
	public final static String CONFIG_PROPERTY = PropertyTransfer.class.getName() + "@config";
	
	private TestStep testStep;
	
	// create local copies since a deleted/changed valuetransfer can be referred to from a result 
	private ValueTransferConfig config;
	private String sourcePath;
	private String sourceType;
	private String targetPath;
	private String name;
	private String targetType;
	private String sourceStep;
	private String targetStep;
	
	private TestStep currentTargetStep;
	private TestStep currentSourceStep;
	private TestStepProperty currentTargetProperty;
	private TestStepProperty currentSourceProperty;
	
	private PropertyChangeSupport propertyChangeSupport = new PropertyChangeSupport( this );
	private StepNameChangeListener stepNameChangeListener = new StepNameChangeListener();
	private PropertyNameChangeListener propertyNameChangeListener = new PropertyNameChangeListener();
	private TestCase testCase;

	private InternalTestSuiteListener testSuiteListener = new InternalTestSuiteListener();

	public PropertyTransfer( TestStep testStep )
	{
		this( testStep, ValueTransferConfig.Factory.newInstance() );
	}
	
	public PropertyTransfer( TestStep testStep, ValueTransferConfig config )
	{
		this.testStep = testStep;
		
		if( testStep != null )
		{
			this.testCase = testStep.getTestCase();
			testCase.getTestSuite().addTestSuiteListener( testSuiteListener );
		}
		
		setConfig( config );
	}
	
	void setConfigOnMove( ValueTransferConfig config )
	{
		this.config = config;
	}
	
	void setConfig(ValueTransferConfig config)
	{
		releaseListeners();
		
		this.config = config;
		
		if( !config.isSetSetNullOnMissingSource() )
		{
			config.setSetNullOnMissingSource( true );
		}
		
		if( !config.isSetTransferTextContent() )
		{
			config.setTransferTextContent( true );
		}
		
		sourceStep = config.getSourceStep();
		if( sourceStep == null )
		{
			sourceStep = getSourceStepName();
			if( sourceStep != null )
				config.setSourceStep( sourceStep );
		}
		
		currentSourceStep = (sourceStep == null || testCase == null ) ? null : testCase.getTestStepByName( sourceStep );

		sourceType = config.getSourceType();
		currentSourceProperty = currentSourceStep == null || sourceType == null ? null : currentSourceStep.getProperty( sourceType );
		
		sourcePath = config.getSourcePath();

		targetStep = config.getTargetStep();
		if( targetStep == null )
		{
			targetStep = getTargetStepName();
			if( targetStep != null )
				config.setTargetStep( targetStep );
		}
		
		currentTargetStep = (targetStep == null || testCase == null ) ? null : testCase.getTestStepByName( targetStep );

		targetType = config.getTargetType();
		currentTargetProperty = currentTargetStep == null || targetType == null ? null : currentTargetStep.getProperty( targetType );
		
		targetPath = config.getTargetPath();
		
		name = config.getName();
		initListeners();
		
		propertyChangeSupport.firePropertyChange( CONFIG_PROPERTY, null, null );
	}
	
	private void initListeners()
	{
		if( currentSourceStep != null )
		{
			currentSourceStep.addPropertyChangeListener( TestStep.NAME_PROPERTY, stepNameChangeListener );
			((WsdlTestStep)currentSourceStep).addTestStepListener( propertyNameChangeListener );
		}

		if( currentTargetStep != null )
		{
			currentTargetStep.addPropertyChangeListener( TestStep.NAME_PROPERTY, stepNameChangeListener );
			((WsdlTestStep)currentTargetStep).addTestStepListener( propertyNameChangeListener );
		}
	}
	
	public void releaseListeners()
	{
		if( currentSourceStep != null )
		{
			currentSourceStep.removePropertyChangeListener( TestStep.NAME_PROPERTY, stepNameChangeListener );
			((WsdlTestStep)currentSourceStep).removeTestStepListener( propertyNameChangeListener );
		}

		if( currentTargetStep != null )
		{
			currentTargetStep.removePropertyChangeListener( TestStep.NAME_PROPERTY, stepNameChangeListener );
			((WsdlTestStep)currentTargetStep).removeTestStepListener( propertyNameChangeListener );
		}
		
		PropertyChangeListener[] listeners = propertyChangeSupport.getPropertyChangeListeners();
		for( PropertyChangeListener listener : listeners )
			propertyChangeSupport.removePropertyChangeListener( listener );
	}
	
	public void release()
	{
		releaseListeners();
		testCase.getTestSuite().removeTestSuiteListener( testSuiteListener );
	}
	
	public ValueTransferConfig getConfig()
	{
		return config;
	}

	public String getSourcePath()
	{
		return sourcePath;
	}
	
	public String getTargetPath()
	{
		return targetPath;
	}
	
	public TestStepProperty getSourceProperty()
	{
		if( sourceType == null )
			return null;
		
		if( currentSourceProperty != null )
			return currentSourceProperty;
		
		TestStep actualSourceStep = getSourceStep();
		return actualSourceStep == null ? null : actualSourceStep.getProperty( sourceType );
	}
	
	public String [] transferProperties(SubmitContext context) throws PropertyTransferException
	{
		TestStepProperty sourceProperty = getSourceProperty();
		TestStepProperty targetProperty = getTargetProperty();
		
		try
		{
			if (sourceProperty == null)
				throw new Exception("Missing source property");
			if (targetProperty == null)
				throw new Exception("Missing target property");
			if (sourceProperty.getValue() == null && !getSetNullOnMissingSource() && !getIgnoreEmpty() )
				throw new Exception("Source property is null");
			
			if (!hasSourcePath() && !hasTargetPath())
			{
				if( !getIgnoreEmpty() || (sourceProperty.getValue() != null && sourceProperty.getValue().length() > 0))
					return transferStringToString(sourceProperty, targetProperty);
			}
			else if (hasSourcePath() && hasTargetPath())
			{
					return transferXPathToXml(sourceProperty, targetProperty, context);
			}
			else if (hasSourcePath() && !hasTargetPath())
			{
				return new String[] { transferXPathToString(sourceProperty, targetProperty, context) };
			}
			else if (!hasSourcePath() && hasTargetPath())
			{
				if( !getIgnoreEmpty() || (sourceProperty.getValue() != null && sourceProperty.getValue().length() > 0))
					return transferStringToXml(sourceProperty, targetProperty, context);
			}
		}
		catch (Exception e)
		{
			throw new PropertyTransferException( e.getMessage(), getSourceStepName(), sourceProperty, 
					getTargetStepName(), targetProperty );
		}		
		
		return new String[0];
	}

	private boolean hasTargetPath()
	{
		String path = getTargetPath();
		return path != null && path.trim().length() > 0 ;
	}

	private boolean hasSourcePath()
	{
		String path = getSourcePath();
		return path != null && path.trim().length() > 0 ;
	}

	protected String[] transferStringToString(TestStepProperty sourceProperty, TestStepProperty targetProperty)
	{
		String value = sourceProperty.getValue();
		targetProperty.setValue( value );
		return new String[]{ value};
	}

	protected String[] transferXPathToXml( TestStepProperty sourceProperty, TestStepProperty targetProperty, SubmitContext context) throws XmlException, Exception
	{
		String sourcePropertyValue = sourceProperty.getValue();
		XmlObject sourceXmlObject = sourcePropertyValue == null ? null : XmlObject.Factory.parse( sourcePropertyValue );
		XmlCursor sourceXml = sourceXmlObject == null ? null : sourceXmlObject.newCursor();
		
		String targetPropertyValue = targetProperty.getValue();
		XmlObject targetXmlObject = XmlObject.Factory.parse( targetPropertyValue );
		XmlCursor targetXml = targetXmlObject.newCursor();

		XmlCursor lastSource = null;
		
		try
		{		
			List<String> result = new ArrayList<String>();

			targetXml.selectPath( PropertyExpansionRequestFilter.expandProperties( context, getTargetPath()) );
			
			if( !targetXml.hasNextSelection() )
				throw new Exception( "Missing match for Target path [" + getTargetPath() + "]" );
			
			if( sourceXml == null )
			{
				if( getSetNullOnMissingSource() )
				{
					while( targetXml.toNextSelection() )
					{
						result.add( setNodeValue( null, targetXml.getDomNode() ));
						if( !getTransferToAll())
							break;
					}
				}
			}
			else if( getUseXQuery() )
			{
				XmlCursor resultCursor = sourceXml.execQuery( PropertyExpansionRequestFilter.expandProperties( context, getSourcePath()) );
				sourceXml.dispose();
				sourceXml = resultCursor;
				
				if( sourceXml.toNextToken() != TokenType.START )
				{
					if( getSetNullOnMissingSource() )
					{
						while( targetXml.toNextSelection() )
						{
							result.add( setNodeValue( null, targetXml.getDomNode() ));
							if( !getTransferToAll())
								break;
						}
					}
					else if( !getIgnoreEmpty() )
						throw new Exception( "Missing match for Source XQuery [" + getSourcePath() + "]" );
				}
				
				boolean hasTarget = targetXml.toNextSelection();
				
				if( hasTarget )
				{
					lastSource = sourceXml.newCursor();
					result.add( transferXmlValue( sourceXml, targetXml ));
				}
			}
			else
			{
				sourceXml.selectPath( PropertyExpansionRequestFilter.expandProperties( context, getSourcePath()) );
				
				if( !sourceXml.hasNextSelection() )
				{
					if( getSetNullOnMissingSource() )
					{
						while( targetXml.toNextSelection() )
						{
							result.add( setNodeValue( null, targetXml.getDomNode() ));
							if( !getTransferToAll())
								break;
						}
					}
					else if( !getIgnoreEmpty() )
						throw new Exception( "Missing match for Source XPath [" + getSourcePath() + "]" );
				}
				else
				{
					boolean hasSource = sourceXml.toNextSelection();
					boolean hasTarget = targetXml.toNextSelection();
					
					while( hasSource && hasTarget )
					{
						if( lastSource != null )
							lastSource.dispose();
						
						lastSource = sourceXml.newCursor();
						result.add( transferXmlValue( sourceXml, targetXml ));
						
						hasSource = sourceXml.toNextSelection();
						hasTarget = targetXml.toNextSelection();
					}
				
					if( getTransferToAll() && !hasSource && hasTarget && lastSource != null )
					{
						while( hasTarget )
						{
							result.add( transferXmlValue( lastSource, targetXml ));
							hasTarget = targetXml.toNextSelection();
						}
					}
				}
			}
			
			if( result.size() > 0 )
			{
				targetProperty.setValue( targetXmlObject.xmlText( new XmlOptions().setSaveAggressiveNamespaces()) );
			}
			
			return result.toArray( new String[result.size()] );
		}
		finally
		{
			sourceXml.dispose();
			targetXml.dispose();
			
			if( lastSource != null )
				lastSource.dispose();
		}
	}

	protected String [] transferStringToXml(TestStepProperty sourceProperty, TestStepProperty targetProperty, SubmitContext context) throws XmlException, Exception
	{
		XmlObject targetXml = XmlObject.Factory.parse( targetProperty.getValue() );
		XmlCursor targetCursor = targetXml.newCursor();
		
		try
		{
			List<String> result = new ArrayList<String>();
			
			targetCursor.selectPath( PropertyExpansionRequestFilter.expandProperties( context, 
					getTargetPath() ));
			
			if( !targetCursor.toNextSelection() )
				throw new Exception( "Missing match for Target XPath [" + getTargetPath() + "]" );
			
			String value = sourceProperty.getValue();
			
			Node targetNode = targetCursor.getDomNode();
			setNodeValue(value, targetNode);
			
			result.add( value );
			
			if( getTransferToAll() )
			{
				while( targetCursor.toNextSelection())
				{
					targetNode = targetCursor.getDomNode();
					setNodeValue(value, targetNode);
					
					result.add( value );
				}
			}
			
			targetProperty.setValue( targetXml.xmlText( new XmlOptions().setSaveAggressiveNamespaces()) );
			
			return result.toArray( new String[result.size()] );
		}
		finally
		{
			targetCursor.dispose();
		}
	}

	private String setNodeValue(String value, Node node) throws Exception
	{
		short targetNodeType = node.getNodeType();
		
		if( targetNodeType == Node.DOCUMENT_FRAGMENT_NODE )
		{
			node = ((DocumentFragment) node ).getFirstChild();
			if( node != null )
				targetNodeType = node.getNodeType();
			else
				throw new Exception( "Missing source value for " + getSourcePropertyName() );
		}
		
		if( targetNodeType == Node.TEXT_NODE || targetNodeType == Node.ATTRIBUTE_NODE )
		{
			node.setNodeValue( value );
		}
		else if( targetNodeType == Node.ELEMENT_NODE )
		{
			XmlUtils.setElementText( (Element) node, value );
		}
		else
		{
			throw new Exception( "Failed to set value to node [" + node.toString() + "] of type [" + targetNodeType + "]" ); 
		}
		
		return value;
	}

	protected String transferXPathToString( TestStepProperty sourceProperty, TestStepProperty targetProperty, SubmitContext context) throws XmlException, Exception
	{
		String sourceValue = sourceProperty.getValue();
		
		if( sourceValue == null )
		{
			if( !getIgnoreEmpty() )
				throw new Exception( "Missing match for Source path [" + getSourcePath() + "]" );
			
			if( getSetNullOnMissingSource() )
				targetProperty.setValue( null );
		}
		
		
		XmlObject sourceXml = sourceValue == null ? null : XmlObject.Factory.parse( sourceValue );
		XmlCursor sourceCursor = sourceValue == null ? null : sourceXml.newCursor();
		
		try
		{
			String value = null;
			
			String xquery = PropertyExpansionRequestFilter.expandProperties( context, getSourcePath());
			if( getUseXQuery() )
			{
				XmlCursor resultCursor = sourceCursor.execQuery( xquery );
				sourceCursor.dispose();
				sourceCursor = resultCursor;
			}
			else
			{
				sourceCursor.selectPath( xquery );
			}
			
			if( !getUseXQuery() && !sourceCursor.toNextSelection() )
			{
				if( !getSetNullOnMissingSource() && !getIgnoreEmpty() )
					throw new Exception( "Missing match for Source path [" + getSourcePath() + "]" );
			}
			else if( getUseXQuery() && sourceCursor.toNextToken() != TokenType.START )
			{
				if( !getSetNullOnMissingSource() && !getIgnoreEmpty() )
					throw new Exception( "Missing match for Source path [" + getSourcePath() + "]" );
			}
			else
			{
				Node sourceNode = sourceCursor.getDomNode();
				short sourceNodeType = sourceNode.getNodeType();
				
				if( sourceNodeType == Node.DOCUMENT_FRAGMENT_NODE )
				{
					sourceNode = ((DocumentFragment) sourceNode ).getFirstChild();
					if( sourceNode != null )
						sourceNodeType = sourceNode.getNodeType();
					else
						throw new Exception( "Missing source value for " + getSourcePropertyName() );
				}
				
				if( sourceNodeType == Node.TEXT_NODE || sourceNodeType == Node.ATTRIBUTE_NODE )
				{
					value = sourceNode.getNodeValue();
				}
				else if( sourceNodeType == Node.ELEMENT_NODE )
				{
					value = XmlUtils.getElementText( (Element) sourceNode );
					if( value == null )
					{
						value = sourceCursor.getObject().xmlText( new XmlOptions().setSaveOuter().setSaveAggressiveNamespaces().setSavePrettyPrint());
					}
				}
			}
			
			if( !getIgnoreEmpty() || (value != null && value.length() >  0) ) 
				targetProperty.setValue( value );
			else
				value = "";
			
			return value;
		}
		finally
		{
			sourceCursor.dispose();
		}
	}

	/**
	 * Method called for transferring between 2 xml properties..
	 */
	
	private String transferXmlValue(XmlCursor source, XmlCursor dest) throws Exception
	{
		// just copy if nodes are of same type
		Node destNode = dest.getDomNode();
		Node sourceNode = source.getDomNode();
		short destNodeType = destNode.getNodeType();
		short sourceNodeType = sourceNode.getNodeType();
		String value = null;
		
		if( sourceNodeType == Node.DOCUMENT_FRAGMENT_NODE )
		{
			sourceNode = ((DocumentFragment) sourceNode ).getFirstChild();
			if( sourceNode != null )
				sourceNodeType = sourceNode.getNodeType();
			else
				throw new Exception( "Missing source value for " + source );
		}
		
		// same type of node?
		if( destNodeType == sourceNodeType )
		{
			if( destNodeType == Node.TEXT_NODE || destNodeType == Node.ATTRIBUTE_NODE )
			{
				value = sourceNode.getNodeValue();
				if( !getIgnoreEmpty() || (value != null && value.length() > 0) )
					destNode.setNodeValue( value );
			}
			else if( config.getTransferTextContent() && destNodeType == Node.ELEMENT_NODE )
			{
				value = XmlUtils.getElementText( (Element) sourceNode);
				if( value == null && sourceNode.getFirstChild() != null )
				{
					value = source.getObject().xmlText( new XmlOptions().setSaveOuter().setSaveAggressiveNamespaces().setSavePrettyPrint() );
					destNode.getParentNode().replaceChild( destNode.getOwnerDocument().importNode( sourceNode, true ), destNode );
				}
				else if( !getIgnoreEmpty() || (value != null && value.length() > 0) )
					XmlUtils.setElementText( (Element) destNode, value );
			}
			else
			{
				destNode = destNode.getParentNode().replaceChild( 
						destNode.getOwnerDocument().importNode( sourceNode, true ), destNode );
				
				value = dest.xmlText();
			}
		}
		// text to attribute?
		else if( (sourceNodeType == Node.TEXT_NODE && destNodeType == Node.ATTRIBUTE_NODE) ||
				(sourceNodeType == Node.ATTRIBUTE_NODE && destNodeType == Node.TEXT_NODE) )
		{
			value = sourceNode.getNodeValue();
			if( !getIgnoreEmpty() || (value != null && value.length() > 0) )
				destNode.setNodeValue( value );
		}
		else if( sourceNodeType == Node.ELEMENT_NODE && 
				destNodeType == Node.ATTRIBUTE_NODE || destNodeType == Node.TEXT_NODE )
		{
			value = XmlUtils.getElementText( (Element) sourceNode );
			if( !getIgnoreEmpty() || (value != null && value.length() > 0) )
				destNode.setNodeValue( value );
		}
		else if( destNodeType == Node.ELEMENT_NODE && 
				sourceNodeType == Node.ATTRIBUTE_NODE || sourceNodeType == Node.TEXT_NODE )
		{
			// hmm.. not sure xmlbeans handles this ok
			value = sourceNode.getNodeValue();
			if( !getIgnoreEmpty() || (value != null && value.length() > 0) )
				XmlUtils.setElementText( (Element) destNode, value );
		}

		return value;
	}
	
	/**
	 * Returns the name of the source property. 
	 */
	
	public String getSourcePropertyName()
	{
		if( sourceType == null )
			return null;
		
		if( currentSourceProperty != null )
			return currentSourceProperty.getName();
		
		TestStep actualSourceStep = getSourceStep();
		if( actualSourceStep == null )
			return sourceType;
		
		TestStepProperty property = actualSourceStep.getProperty( sourceType );
		return property == null ? sourceType : property.getName();
	}
	
	public void setSourcePropertyName( String name )
	{
		String old = getSourcePropertyName();
		
		// check for change
		if( (name == null && old == null) || (name != null && old != null && name.equals( old ))) return;

		// update
		sourceType = name;
		config.setSourceType( name );
		
		// update actual property
		TestStep sourceStep2 = getSourceStep();
		currentSourceProperty = sourceStep2 != null && sourceType != null ? 
				sourceStep2.getProperty( sourceType ) : null;
		
		// notify!
		propertyChangeSupport.firePropertyChange( SOURCE_TYPE_PROPERTY, old, name );
	}
	
	public TestStepProperty getTargetProperty()
	{
		if( targetType == null )
			return null;
		
		if( currentTargetProperty != null )
			return currentTargetProperty;
		
		TestStep actualTargetStep = getTargetStep();
		return actualTargetStep == null ? null : actualTargetStep.getProperty( targetType );
	}
	
	public String getTargetPropertyName()
	{
		if( targetType == null )
			return null;
		
		if( currentTargetProperty != null )
			return currentTargetProperty.getName();
		
		TestStep actualTargetStep = getTargetStep();
		TestStepProperty property = actualTargetStep == null ? null : actualTargetStep.getProperty( targetType );
		return actualTargetStep == null || property == null ? targetType: property.getName();
	}
	
	public void setTargetPropertyName( String name )
	{
		String old = getTargetPropertyName();
		
		// check for change
		if( (name == null && old == null) || (name != null && old != null && name.equals( old ))) return;

		// update
		targetType = name;
		config.setTargetType( name );
		
		// update actual property
		TestStep targetStep2 = getTargetStep();
		
		currentTargetProperty = targetStep2 != null && targetType != null ? 
				targetStep2.getProperty( targetType ) : null;
		
		// notify!
		propertyChangeSupport.firePropertyChange( TARGET_TYPE_PROPERTY, old, name );
	}
	
	public String getName()
	{
		return config.getName();
	}
	
	public void setSourcePath( String path )
	{
		String old = sourcePath;
		sourcePath = path;
		config.setSourcePath( path );
		propertyChangeSupport.firePropertyChange( SOURCE_PATH_PROPERTY, old, path );
	}
	
	public void setTargetPath( String path )
	{
		String old = targetPath;
		targetPath = path;
		config.setTargetPath( path );
		propertyChangeSupport.firePropertyChange( TARGET_PATH_PROPERTY, old, path );
	}
	
	public void setName( String name )
	{
		String old = this.name;
		this.name = name;
		config.setName( name );
		propertyChangeSupport.firePropertyChange( NAME_PROPERTY, old, name );
	}

	public TestStep getSourceStep()
	{
		String sourceStepName = getSourceStepName();
		return sourceStepName == null ? null : testCase.getTestStepByName( sourceStepName );
	}

	public String getSourceStepName()
	{
		if( sourceStep != null )
			return sourceStep;
		
		if( testCase == null )
			return null;
		
		TestStep step = testCase.findPreviousStepOfType( this.testStep, WsdlTestRequestStep.class );
		return step == null ? null : step.getName();
	}

	public void setSourceStepName(String sourceStep)
	{
		String old = getSourceStepName();
		
		//	 check for change
		if( (sourceStep == null && old == null) || (sourceStep != null && old != null && sourceStep.equals( old ))) return;
		
		if( sourceStep == null )
			log.debug( "Setting sourceStep for transfer [" + getName() + "] to null" );
		
		this.sourceStep = sourceStep;
		config.setSourceStep( sourceStep );
		
		if( currentSourceStep != null )
		{
			currentSourceStep.removePropertyChangeListener( stepNameChangeListener );
			((WsdlTestStep)currentSourceStep).removeTestStepListener( propertyNameChangeListener );
		}
		
		currentSourceStep = (WsdlTestStep) (sourceStep == null ? null : testStep.getTestCase().getTestStepByName( sourceStep ));
		if( currentSourceStep != null )
		{
			currentSourceStep.addPropertyChangeListener( TestStep.NAME_PROPERTY, stepNameChangeListener );
			((WsdlTestStep)currentSourceStep).addTestStepListener( propertyNameChangeListener );
		}
		
		propertyChangeSupport.firePropertyChange( SOURCE_STEP_PROPERTY, old, sourceStep );
		setSourcePropertyName( null );
	}
	
	public TestStep getTargetStep()
	{
		String targetStepName = getTargetStepName();
		return targetStepName == null ? null : testCase.getTestStepByName( targetStepName );
	}

	public String getTargetStepName()
	{
		if( targetStep != null )
			return targetStep;
		
		if( testCase == null )
			return null;
		
		TestStep step = testCase.findNextStepOfType( this.testStep, WsdlTestRequestStep.class );
		return step == null ? null : step.getName();
	}

	public void setTargetStepName(String targetStep)
	{
		String old = getTargetStepName();
		
		//	 check for change
		if( (targetStep == null && old == null) || (targetStep != null && old != null && targetStep.equals( old ))) return;
		
		if( targetStep == null )
			log.debug( "Setting targetStep for transfer [" + getName() + "] to null" );
		
		this.targetStep = targetStep;
		config.setTargetStep( targetStep );
		
		if( currentTargetStep != null )
		{
			currentTargetStep.removePropertyChangeListener( stepNameChangeListener );
			((WsdlTestStep)currentTargetStep).removeTestStepListener( propertyNameChangeListener );
		}
		
		currentTargetStep = (WsdlTestStep) (targetStep == null ? null : testStep.getTestCase().getTestStepByName( targetStep ));
		if( currentTargetStep != null )
		{
			currentTargetStep.addPropertyChangeListener( TestStep.NAME_PROPERTY, stepNameChangeListener );
			((WsdlTestStep)currentTargetStep).addTestStepListener( propertyNameChangeListener );
		}
		
		propertyChangeSupport.firePropertyChange( TARGET_STEP_PROPERTY, old, targetStep );
		setTargetPropertyName( null );
	}

	public void addPropertyChangeListener(String propertyName, PropertyChangeListener listener)
	{
		propertyChangeSupport.addPropertyChangeListener( propertyName, listener );
	}

	public void addPropertyChangeListener(PropertyChangeListener listener)
	{
		propertyChangeSupport.addPropertyChangeListener( listener );
	}

	public void removePropertyChangeListener(PropertyChangeListener listener)
	{
		propertyChangeSupport.removePropertyChangeListener( listener );
	}

	public void removePropertyChangeListener(String propertyName, PropertyChangeListener listener)
	{
		propertyChangeSupport.removePropertyChangeListener( propertyName, listener );
	}
	
	public boolean getFailOnError()
	{
		return config.getFailOnError();
	}
	
	public void setFailOnError( boolean failOnError )
	{
		config.setFailOnError( failOnError );
	}
	
	public boolean getTransferToAll()
	{
		return config.getTransferToAll();
	}
	
	public void setTransferToAll( boolean transferToAll )
	{
		config.setTransferToAll( transferToAll );
	}
	
	public boolean getUseXQuery()
	{
		return config.getUseXQuery();
	}
	
	public void setUseXQuery( boolean useXQuery )
	{
		config.setUseXQuery( useXQuery );
	}
	
	public boolean getIgnoreEmpty()
	{
		return config.getIgnoreEmpty();
	}
	
	public void setIgnoreEmpty( boolean ignoreEmpty )
	{
		config.setIgnoreEmpty( ignoreEmpty );
	}
	
	public boolean getSetNullOnMissingSource()
	{
		return config.getSetNullOnMissingSource();
	}
	
	public void setSetNullOnMissingSource( boolean setNullOnMissingSource )
	{
		config.setSetNullOnMissingSource( setNullOnMissingSource );
	}
	
	public boolean getTransferTextContent()
	{
		return config.getTransferTextContent();
	}
	
	public void setTransferTextContent( boolean transferTextContent )
	{
		config.setTransferTextContent( transferTextContent );
	}
	
	private final class InternalTestSuiteListener extends TestSuiteListenerAdapter
	{
		public void testStepRemoved(TestStep testStep, int index)
		{
			if( testStep.getTestCase() == testCase )
			{
				String stepName = testStep.getName();
				if( stepName.equals( sourceStep ) )
				{
					setSourceStepName( null );
				}
			
				if( stepName.equals( targetStep ) )
				{
					setTargetStepName( null );
				}
			}
		}
	}

	/**
	 * Handle changes to source/target testStep names
	 * 
	 * @author Ole.Matzura
	 */
	
	private class StepNameChangeListener implements PropertyChangeListener
	{
		public void propertyChange(PropertyChangeEvent evt)
		{
			String oldName = (String) evt.getOldValue();
			String newValue = (String) evt.getNewValue();
			
			if( newValue == null )
			{
				log.error( "Tried to change stepname to null!" );
				Thread.dumpStack();
				return;
			}
			
			if( oldName.equals( sourceStep ))
			{
				sourceStep = newValue;
				config.setSourceStep( sourceStep );
				propertyChangeSupport.firePropertyChange( SOURCE_STEP_PROPERTY, oldName, sourceStep );
			}
			
			if( oldName.equals( targetStep ))
			{
				targetStep = newValue;
				config.setTargetStep( targetStep );
				propertyChangeSupport.firePropertyChange( TARGET_STEP_PROPERTY, oldName, targetStep );
			}
		}
	}
	
	/**
	 * Handle changes to source/target property names
	 * 
	 * @author Ole.Matzura
	 */
	
	private class PropertyNameChangeListener extends WsdlTestStepListenerAdapter
	{
		public void propertyRenamed(String oldName, String newName)
		{
			if( oldName.equals( sourceType ))
			{
				sourceType = newName;
				config.setSourceType( sourceType );
				propertyChangeSupport.firePropertyChange( SOURCE_TYPE_PROPERTY, oldName, sourceType );
			}
			
			if( oldName.equals( targetType ))
			{
				targetType = newName;
				config.setTargetType( targetType );
				propertyChangeSupport.firePropertyChange( TARGET_TYPE_PROPERTY, oldName, targetType );
			}
		}

		public void propertyRemoved(String name)
		{
			if( name.equals( sourceType ))
			{
				log.warn( "source property for transfer [" + getName() + "] in teststep [" + 
						testStep.getName() + "] set to null, was [" + name + "]" );  
				
				setSourcePropertyName( null );
			}

			if( name.equals( targetType ))
			{
				log.warn( "target property for transfer [" + getName() + "] in teststep [" + 
						testStep.getName() + "] set to null, was [" + name + "]" );  
				
				setTargetPropertyName( null );
			}
		}
	}
}